import win32.windows;
import win32.windef;
import win32.commctrl;
import win32.vsstyle;
import win32.uxtheme;
import win32frame.WindowObject;
import util.BitMap;
import std.stdio;
import std.string;
import std.utf;

COLORREF RGB(uint r, uint g, uint b) {
  return cast(COLORREF)(r | (g << 8) | (b << 16));
}

class MyButton : WindowObject {
  HFONT    labelfont ;
  BitMap   bitmap ;
  bool     push ;
  bool     focus ;
  bool     selected ;
  bool     FlatMode ;
  bool     withBG ;
  uint     selcol ;
  uint     focuscol ;
  uint     strcol ;
  
  HTHEME hThemeBTN ;
  HTHEME hThemeBG  ;

  this(string name) {
    winc.style          = 0;
    this.setStyle(WS_CHILD | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_VISIBLE) ;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.lpszMenuName   = null;
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    winc.lpszClassName  = toUTFz!(wchar*)("MyButton") ;
    super(name) ;
    this.registwinc() ;
    selcol   =RGB(192,192,255) ;
    focuscol =RGB(255,128,128) ;
    strcol   =RGB(  0,  0,  0) ;
  }
  
  override int setFont(HFONT hfont){
    this.labelfont=hfont ;
    if( getHWND()!=null ){
      InvalidateRect(getHWND() , null , true);
    }
    return(0) ;
  }

  int setSelectColor(uint col){
    selcol=col ;
    return(0) ;
  }

  int setSelectState(bool st){
    selected = st ;
    if( getHWND()!=null ){
      InvalidateRect(getHWND() , null , true);
    }
    return(0) ;
  }

  bool getSelectState(){
    return(selected) ;
  }

  int setBitMap(BitMap bmp){
    this.bitmap = bmp ;
    return(0) ;
  }

  uint setStringColor(int r, int g, int b){
    uint oldcol=strcol ;
    strcol=RGB(r,g,b) ;
    if( getHWND()!=null ){
      InvalidateRect(getHWND() , null , true);
    }
    return(oldcol) ;
  }

  uint setStringColor(uint color){
    uint oldcol=strcol ;
    strcol=color ;
    if( getHWND()!=null ){
      InvalidateRect(getHWND() , null , true);
    }
    return(oldcol) ;
  }

  int setFlatMode(bool flatmd, bool BG=false){
    FlatMode=flatmd ;
    withBG=BG ;
    if( isThemeSupported() == true ){
      if( getHWND()!=null){
	if( hThemeBTN!=null ) CloseThemeData(hThemeBTN) ;
	hThemeBTN = OpenThemeData(getHWND(), toUTFz!(wchar*)(FlatMode==true ? VSCLASS_TOOLBAR : VSCLASS_BUTTON)) ;
	if( withBG==true ){
	  if( hThemeBG !=null ) CloseThemeData(hThemeBG) ;
	  hThemeBG = OpenThemeData(getHWND(), toUTFz!(wchar*)(VSCLASS_REBAR)) ;
	}
      }
    }
    return(0) ;
  }

  LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch (msg) {
    case WM_DESTROY:
      if( hThemeBTN!=null ) CloseThemeData(hThemeBTN) ;
      if( hThemeBG !=null ) CloseThemeData(hThemeBG ) ;
      break ;
      
    case WM_CREATE :
      if( isThemeSupported() == true ){
	hThemeBTN = OpenThemeData(getHWND(), toUTFz!(wchar*)(FlatMode==true ? VSCLASS_TOOLBAR : VSCLASS_BUTTON)) ;
	if( withBG==true ){
	  hThemeBG = OpenThemeData(getHWND(), toUTFz!(wchar*)(VSCLASS_REBAR)) ;
	}
      }
      break ;
      
    case WM_PAINT:
      HDC hdc, hdcbmp;
      PAINTSTRUCT ps ;
      HBRUSH  COLBASE ;
      HPEN    COLHILIGHT, COLLIGHT, COLSHADE, COLDKSHADE ;
      HPEN    COLFOCUS ;
      HBITMAP hbmp ;
      uint    bgcol ;

      if( selected==true ){
	uint r = (selcol&0x000000ff)     ;
	uint g = (selcol&0x0000ff00)>>8  ;
	uint b = (selcol&0x00ff0000)>>16 ;
	bgcol     = selcol ;
	COLBASE   = CreateSolidBrush(bgcol) ;
	COLHILIGHT= CreatePen(PS_SOLID, 1, RGB(r/2+127  ,g/2+127  ,b/2+127  )) ;
	COLLIGHT  = CreatePen(PS_SOLID, 1, RGB(r*3/4+63 ,g*3/4+63 ,b*3/4+63 )) ;
	COLSHADE  = CreatePen(PS_SOLID, 1, RGB(r*3/4    ,g*3/4    ,b*3/4    )) ;
	COLDKSHADE= CreatePen(PS_SOLID, 1, RGB(r/2      ,g/2      ,b/2      )) ;
      }else{
	bgcol     = GetSysColor(COLOR_3DFACE) ;
	COLBASE   = CreateSolidBrush(bgcol) ;
	COLHILIGHT= CreatePen(PS_SOLID, 1, GetSysColor(COLOR_3DHILIGHT )) ;
	COLLIGHT  = CreatePen(PS_SOLID, 1, GetSysColor(COLOR_3DLIGHT ))   ;
	COLSHADE  = CreatePen(PS_SOLID, 1, GetSysColor(COLOR_3DSHADOW))   ;
	COLDKSHADE= CreatePen(PS_SOLID, 1, GetSysColor(COLOR_3DDKSHADOW)) ;
      }
      COLFOCUS = CreatePen(PS_SOLID, 1, focuscol)   ;

      hdc = BeginPaint(hwnd, &ps);
      
      if( hThemeBTN!=null ){
	RECT rc ;
	GetClientRect( hwnd, &rc ) ;
	int iStateId = (this.style & BS_DEFPUSHBUTTON)!=0 ? PBS_DEFAULTED : push==true ? PBS_PRESSED : (focus==true||selected==true) ? PBS_HOT : PBS_NORMAL ;
	if( bitmap !is null ){
	  int imgw = bitmap.getWidth()  ;
	  int imgh = bitmap.getHeight() ;
	  int px = (this.style & BS_CENTER )==BS_CENTER || (this.style & BS_CENTER )==0 ? (winW-imgw)/2+1 : (this.style & BS_RIGHT )==BS_RIGHT ? (winW-imgw) : 0 ;
	  int py = (this.style & BS_VCENTER)==BS_VCENTER|| (this.style & BS_VCENTER)==0 ? (winH-imgh)/2+1 : (this.style & BS_BOTTOM)==BS_BOTTOM? (winH-imgh) : 0 ;
	  hdcbmp = CreateCompatibleDC(hdc) ;
	  
	  uint *bmpp ;
	  BITMAPINFO bi ;
	  GetClientRect( hwnd, &rc ) ;
	  bi.bmiHeader.biSize             = BITMAPINFOHEADER.sizeof;
	  bi.bmiHeader.biWidth            =  (rc.right-rc.left+1) ;
	  bi.bmiHeader.biHeight           = -(rc.bottom-rc.top+1) ;
	  bi.bmiHeader.biPlanes           = 1;
	  bi.bmiHeader.biBitCount         = 32;
	  bi.bmiHeader.biCompression      = BI_RGB;
	  bi.bmiHeader.biSizeImage        = 0;
	  bi.bmiHeader.biXPelsPerMeter    = 0;
	  bi.bmiHeader.biYPelsPerMeter    = 0;
	  bi.bmiHeader.biClrUsed          = 1;
	  bi.bmiHeader.biClrImportant     = 0;
	  
	  hbmp=CreateDIBSection( hdc, &bi, DIB_RGB_COLORS, cast(void**)(&bmpp), null, 0 );
	  SelectObject( hdcbmp, hbmp ) ;
	  if( hThemeBG!=null ){
	    RECT exprc = rc ; exprc.left-=2 ; exprc.right+=2 ;
	    DrawThemeBackground(hThemeBG, hdcbmp, RP_BACKGROUND,  0, &exprc, null);
	  }else{
	    FillRect(hdcbmp, &rc, COLBASE) ;
	  }
	  DrawThemeBackground(hThemeBTN, hdcbmp,  BP_PUSHBUTTON,  iStateId, &rc, null);
	  if( push==true ){ px++ ; py++ ; }
	  bitmap.bitblt(bmpp, bi.bmiHeader.biWidth, -bi.bmiHeader.biHeight, px, py) ;
	  BitBlt( hdc, 0,0, bi.bmiHeader.biWidth, -bi.bmiHeader.biHeight, hdcbmp, 0,0, SRCCOPY ) ;
	}else{
	  SIZE labelsize ;
	  wchar[] label=(toUTF16(getWindowText())).dup ;
	  
	  if( hThemeBG!=null ){
	    RECT exprc = rc ; exprc.left-=2 ; exprc.right+=2 ;
	    DrawThemeBackground(hThemeBG, hdc, RP_BACKGROUND,  0, &exprc, null);
	  }else{
	    FillRect(hdc, &rc, COLBASE) ;
	  }
	  DrawThemeBackground(hThemeBTN, hdc, BP_PUSHBUTTON, iStateId, &rc, null);
	  SetBkMode(hdc , TRANSPARENT);
	  
	  GetTextExtentPoint32(hdc , cast(wchar*)(label) , cast(int)(label.length) , &labelsize);
	  SelectObject(hdc,labelfont) ;
	  SetTextColor(hdc,strcol) ;
	  int d  = (this.style & BS_DEFPUSHBUTTON)!=0 ? 1 : 0 ;
	  int px = (this.style & BS_CENTER )==BS_CENTER || (this.style & BS_CENTER )==0 ? (winW-labelsize.cx)/2+1 : (this.style & BS_RIGHT )==BS_RIGHT ? (winW-labelsize.cx) : d+3 ;
	  int py = (this.style & BS_VCENTER)==BS_VCENTER|| (this.style & BS_VCENTER)==0 ? (winH-labelsize.cy)/2+1 : (this.style & BS_BOTTOM)==BS_BOTTOM? (winH-labelsize.cy) : d+3 ;
	  if( push==true ){ px++ ; py++ ; }
	  TextOut(hdc,px,py,cast(wchar*)(label), cast(int)(label.length)) ;
	}
	
      }else{
	RECT rc ;
	GetClientRect( hwnd, &rc ) ;

	if( bitmap !is null ){
	  int imgw = bitmap.getWidth()  ;
	  int imgh = bitmap.getHeight() ;
	  int px = (this.style & BS_CENTER )==BS_CENTER || (this.style & BS_CENTER )==0 ? (winW-imgw)/2+1 : (this.style & BS_RIGHT )==BS_RIGHT ? (winW-imgw) : 0 ;
	  int py = (this.style & BS_VCENTER)==BS_VCENTER|| (this.style & BS_VCENTER)==0 ? (winH-imgh)/2+1 : (this.style & BS_BOTTOM)==BS_BOTTOM? (winH-imgh) : 0 ;
	  hdcbmp = CreateCompatibleDC(hdc) ;

	  uint *bmpp ;
	  BITMAPINFO bi ;
	  GetClientRect( hwnd, &rc ) ;
	  bi.bmiHeader.biSize             = BITMAPINFOHEADER.sizeof;
	  bi.bmiHeader.biWidth            =  (rc.right-rc.left+1) ;
	  bi.bmiHeader.biHeight           = -(rc.bottom-rc.top+1) ;
	  bi.bmiHeader.biPlanes           = 1;
	  bi.bmiHeader.biBitCount         = 32;
	  bi.bmiHeader.biCompression      = BI_RGB;
	  bi.bmiHeader.biSizeImage        = 0;
	  bi.bmiHeader.biXPelsPerMeter    = 0;
	  bi.bmiHeader.biYPelsPerMeter    = 0;
	  bi.bmiHeader.biClrUsed          = 1;
	  bi.bmiHeader.biClrImportant     = 0;
	  
	  hbmp=CreateDIBSection( hdc, &bi, DIB_RGB_COLORS, cast(void**)(&bmpp), null, 0 );
	  SelectObject( hdcbmp, hbmp ) ;
	  FillRect(hdcbmp, &rc, COLBASE) ;
	  if( push==true ){ px++ ; py++ ; }
	  bitmap.bitblt(bmpp, bi.bmiHeader.biWidth, -bi.bmiHeader.biHeight, px, py) ;
	  BitBlt( hdc, 0,0, bi.bmiHeader.biWidth, -bi.bmiHeader.biHeight, hdcbmp, 0,0, SRCCOPY ) ;
	}else{
	  FillRect(hdc, &rc, COLBASE) ;
	}
	int d = (this.style & BS_DEFPUSHBUTTON)!=0 ? 1 : 0 ;
	if( push==false ){
	  if( FlatMode==false ){
	    SelectObject(hdc, COLHILIGHT) ;
	    MoveToEx(hdc,      d, winH-(d+1), null) ;
	    LineTo(  hdc,      d,      d) ;
	    LineTo(  hdc, winW-d,      d) ;
	    SelectObject(hdc, COLLIGHT) ;
	    MoveToEx(hdc,      (d+1), winH-(d+2), null) ;
	    LineTo(  hdc,      (d+1),      (d+1)) ;
	    LineTo(  hdc, winW-(d+1),      (d+1)) ;
	    SelectObject(hdc, COLDKSHADE) ;
	    MoveToEx(hdc,      (d+1), winH-(d+1), null) ;
	    LineTo(  hdc, winW-(d+1), winH-(d+1)) ;
	    LineTo(  hdc, winW-(d+1),          d) ;
	    SelectObject(hdc, COLSHADE) ;
	    MoveToEx(hdc,      (d+2), winH-(d+2), null) ;
	    LineTo(  hdc, winW-(d+2), winH-(d+2)) ;
	    LineTo(  hdc, winW-(d+2),      (d+1)) ;
	  }
	}else{
	  SelectObject(hdc, COLDKSHADE) ;
	  MoveToEx(hdc,      d, winH-(d+1), null) ;
	  LineTo(  hdc,      d,      d) ;
	  LineTo(  hdc, winW-d,      d) ;
	  SelectObject(hdc, COLSHADE) ;
	  MoveToEx(hdc,      (d+1), winH-(d+2), null) ;
	  LineTo(  hdc,      (d+1),      (d+1)) ;
	  LineTo(  hdc, winW-(d+1),      (d+1)) ;
	  SelectObject(hdc, COLHILIGHT) ;
	  MoveToEx(hdc,      (d+1), winH-(d+1), null) ;
	  LineTo(  hdc, winW-(d+1), winH-(d+1)) ;
	  LineTo(  hdc, winW-(d+1),      d) ;
	  SelectObject(hdc, COLLIGHT) ;
	  MoveToEx(hdc,      (d+2), winH-(d+2), null) ;
	  LineTo(  hdc, winW-(d+2), winH-(d+2)) ;
	  LineTo(  hdc, winW-(d+2),      (d+1)) ;
	}
	if( focus==true ){
	  SelectObject(hdc, COLFOCUS) ;
	  MoveToEx(hdc,      0,      0, null) ;
	  LineTo(  hdc, winW-1,      0) ;
	  LineTo(  hdc, winW-1, winH-1) ;
	  LineTo(  hdc,      0, winH-1) ;
	  LineTo(  hdc,      0,      0) ;
	}
	
	SetBkMode(hdc , TRANSPARENT);
	if( bitmap is null ){
	  SIZE labelsize ;
	  wchar[] label=(toUTF16(getWindowText())).dup ;
	  
	  GetTextExtentPoint32(hdc , cast(wchar*)(label) , cast(int)(label.length) , &labelsize);
	  SelectObject(hdc,labelfont) ;
	  SetTextColor(hdc,strcol) ;
	  int px = (this.style & BS_CENTER )==BS_CENTER || (this.style & BS_CENTER )==0 ? (winW-labelsize.cx)/2+1 : (this.style & BS_RIGHT )==BS_RIGHT ? (winW-labelsize.cx) : d+3 ;
	  int py = (this.style & BS_VCENTER)==BS_VCENTER|| (this.style & BS_VCENTER)==0 ? (winH-labelsize.cy)/2+1 : (this.style & BS_BOTTOM)==BS_BOTTOM? (winH-labelsize.cy) : d+3 ;
	  if( push==true ){ px++ ; py++ ; }
	  TextOut(hdc,px,py,cast(wchar*)(label), cast(int)(label.length)) ;
	}
      }

      EndPaint(hwnd, &ps);
      
      DeleteObject(COLBASE ) ;
      DeleteObject(COLHILIGHT) ;
      DeleteObject(COLLIGHT) ;
      DeleteObject(COLSHADE) ;
      DeleteObject(COLDKSHADE) ;
      DeleteObject(COLFOCUS) ;
      if( bitmap !is null ){
	DeleteObject(hbmp) ;
	DeleteDC(hdcbmp) ;
      }
      break ;

    case WM_SIZE:
      RedrawWindow( hwnd, null, null, RDW_INVALIDATE) ;
      break;

    case WM_MOUSELEAVE:
      focus = false ;
      RedrawWindow( hwnd, null, null, RDW_INVALIDATE) ;
      break ;

    case WM_MOUSEMOVE:
      if( focus==false ){
        TRACKMOUSEEVENT trackmsevent ;
        trackmsevent.cbSize  = TRACKMOUSEEVENT.sizeof ;
        trackmsevent.dwFlags = TME_LEAVE ;
        trackmsevent.hwndTrack = hwnd ;
        TrackMouseEvent( &trackmsevent ) ;
	focus = true ;
      }
      RedrawWindow( hwnd, null, null, RDW_INVALIDATE) ;
      break ;
      
    case WM_LBUTTONDOWN:
      push=true ;
      SetCapture( hwnd ) ;
      InvalidateRect(hwnd , null , false);
      SetFocus(hwnd) ;
      break ;
      
    case WM_LBUTTONUP:
      push=false ;
      InvalidateRect(hwnd , null , false);
      ReleaseCapture() ;
      focus = false ;
      sendmsgtoparent( WM_COMMAND, cast(int)(getHMENU()), 0) ;
      break ;

    case WM_THEMECHANGED:      
      if( isThemeSupported() == true ){
	if( hThemeBTN!=null ) CloseThemeData(hThemeBTN) ;
	hThemeBTN = OpenThemeData(getHWND(), toUTFz!(wchar*)(FlatMode==true ? VSCLASS_TOOLBAR : VSCLASS_BUTTON)) ;
	if( withBG==true ){
	  if( hThemeBG !=null ) CloseThemeData(hThemeBG) ;
	  hThemeBG = OpenThemeData(getHWND(), toUTFz!(wchar*)(VSCLASS_REBAR)) ;
	}
      }
      break ;

    default :
      return( DefWindowProc(hwnd , msg , wp , lp) );
    }
    return(0);
  }

}
