import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.glut;
import std.stdio;
import std.string;
import std.utf;
import std.math;
import std.format;

import OpenGLBase;

const LBUTTONDOWN   = (0x0001) ;
const LBUTTONUP     = (0x0002) ;
const LBUTTONDBLCLK = (0x0004) ;
const RBUTTONDOWN   = (0x0001) ;
const RBUTTONUP     = (0x0002) ;
const RBUTTONDBLCLK = (0x0004) ;
const ALT_KEY       = (0x0001) ;
const CNTL_KEY      = (0x0002) ;
const SHIFT_KEY     = (0x0004) ;

struct Camera{
  double fovy=30.0 ;
  double ex= 0.0, ey= 0.0, ez=10.0 ;
  double cx= 0.0, cy= 0.0, cz= 0.0 ;
  double rx=45.0, ry=45.0, rz=45.0 ;
  double px= 0.0, py= 0.0, pz= 0.0 ;
  double zoom=1.0 ;
} ;

class OpenGLFrameTest : OpenGLBase {
  Camera camera ;
  int win_w  , win_h ;
  int mouse_Button, startx, starty ;
  int solidobj=0 ;
  
  GLfloat[] mat_diffuse ;
  GLfloat[] mat_ambient ;
    
  this(string name, string ClassName) {
    super(name,ClassName) ;
    mat_diffuse = new GLfloat[4] ;
    mat_diffuse[0]=0.8 ;
    mat_diffuse[1]=0.8 ;
    mat_diffuse[2]=0.8 ;
    mat_diffuse[3]=1.0 ;
    mat_ambient = new GLfloat[4] ;
    mat_ambient[0]=0.5 ;
    mat_ambient[1]=0.5 ;
    mat_ambient[2]=0.5 ;
    mat_ambient[3]=1.0 ;

  }

  int wm_create(){
    glClearColor(0.0,0.0,1.0,0.5) ;

    glMaterialfv(GL_FRONT, GL_DIFFUSE, &mat_diffuse[0]);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glLightfv(GL_LIGHT0, GL_AMBIENT, &mat_ambient[0] ) ;
    
    return(0) ;
  }

  int wm_paint(){
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glLoadIdentity();

    glViewport(0, 0, win_w, win_h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(camera.fovy, cast(double)(win_w)/cast(double)(win_h), 1.0, 100.0);
    gluLookAt(camera.ex, camera.ey, camera.ez,
	      camera.cx, camera.cy, camera.cz,
	      0.0,1.0, 0.0) ;
    glMatrixMode(GL_MODELVIEW);

    glClearColor(0.0,0.0,1.0,0.5) ;

    glMaterialfv(GL_FRONT, GL_DIFFUSE, &mat_diffuse[0]);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glLightfv(GL_LIGHT0, GL_AMBIENT, &mat_ambient[0] ) ;
    
    glTranslated( camera.px , camera.py , camera.pz);
    glRotated( camera.rx , 1.0, 0.0, 0.0 );
    glRotated( camera.ry , 0.0, 1.0, 0.0 );
    glRotated( camera.rz , 0.0, 0.0, 1.0 );

    switch( solidobj ){
    case 1 : glutSolidSphere(1.0, 32, 32)     ; break ;
    case 2 : glutSolidCube( 2.0)              ; break ;
    case 3 : glutSolidCone( 1.0, 2.0, 32, 32) ; break ;
    case 4 : glutSolidTorus(0.4, 1.0, 32, 32) ; break ;
    default :
      glutSolidTeapot(1.0);
    }
    
    //    GLUquadricObj *quad = gluNewQuadric();
    //    gluQuadricDrawStyle(quad, GLU_FILL);
    //    gluCylinder(quad, 1.0, 1.0, 1.0, 64, 1);
    //    gluDeleteQuadric(quad);

    glDisable( GL_LIGHTING );
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    gluOrtho2D( 0.0, win_w, 0.0, win_h  );
    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
    glColor3f( 1.0, 1.0, 1.0 ) ;
    
    glut_print(4,30,"Rotate : Alt+Mouse'L'drag") ;
    glut_print(4,17,"Zoom   : Alt+Cntl+Mouse'L'drag") ;
    glut_print(4, 4,"Shift  : Alt+Shift+Mouse'L'drag") ;

    glFinish();
    return(0) ;
  }

  void glut_print(int x, int y, ...){
    void delegate(dchar) putc = &glut_putc ;
    glRasterPos2i(x,y) ;
    std.format.doFormat( putc, _arguments, _argptr ) ;
  }
  
  void glut_putc(dchar c){
    glutBitmapCharacter(cast(void*)(GLUT_BITMAP_8_BY_13), c);
  }
  
  int wm_size(int w, int h){
    win_w=w ; win_h=h ;
    
    glViewport(0, 0, win_w, win_h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(camera.fovy, cast(double)(win_w)/cast(double)(win_h), 1.0, 100.0);
    gluLookAt(camera.ex, camera.ey, camera.ez,
	      camera.cx, camera.cy, camera.cz,
	      0.0,1.0, 0.0) ;
    glMatrixMode(GL_MODELVIEW);
    win_w=w ;
    win_h=h ;
    return(0) ;
  }

  int mouse( int acs_stat, int mb, int x, int y ){
    int alt_stat ;
    int shift_stat ;
    int cntl_stat ;
    
    alt_stat  = acs_stat & ALT_KEY   ;
    shift_stat= acs_stat & SHIFT_KEY ;
    cntl_stat = acs_stat & CNTL_KEY ;

    if( (mb & LBUTTONDOWN)!=0 ) mouse_Button |= 1 ;
    if( (mb & RBUTTONDOWN)!=0 ) mouse_Button |= 2 ;
    if( (mb & LBUTTONUP  )!=0 ) mouse_Button &= ~(1) ;
    if( (mb & RBUTTONUP  )!=0 ) mouse_Button &= ~(2) ;
    
    if( mouse_Button==0 ){
      startx=x ;
      starty=y ;
      return(0) ;
    }
    if( alt_stat!=0 ){
      if(  shift_stat==0 && cntl_stat==0 ){
        camera.ry += startx-x ;
        if( camera.ry<0.0   ) camera.ry+=360.0 ;
        if( camera.ry>360.0 ) camera.ry-=360.0 ;
        camera.rx -= starty-y ;
        if( camera.rx<0.0   ) camera.rx+=360.0 ;
        if( camera.rx>360.0 ) camera.rx-=360.0 ;
        startx=x ;
        starty=y ;
        return(1) ;
      }
      
      if( shift_stat==0 && cntl_stat!=0 ){
        camera.ez += (starty-y)/10.0 ;
        if( camera.ez>  100 ) camera.ez=  100 ;
        if( camera.ez<  0.0 ) camera.ez=  0.0 ;
        startx=x ;
        starty=y ;
        return(1) ;
      }
      
      if( shift_stat!=0 && cntl_stat==0 ){
	double dwin_w = cast(double)(win_w) ;
	double dwin_h = cast(double)(win_h) ;
	camera.px += (camera.ez*tan(camera.fovy/2.0*PI/180.0) * (cast(double)(x     ) - (dwin_w/2.0))/(dwin_h/2.0)) -
	             (camera.ez*tan(camera.fovy/2.0*PI/180.0) * (cast(double)(startx) - (dwin_w/2.0))/(dwin_h/2.0)) ;
	camera.py +=(-camera.ez*tan(camera.fovy/2.0*PI/180.0) * (cast(double)(y     ) - (dwin_h/2.0))/(dwin_h/2.0)) -
	            (-camera.ez*tan(camera.fovy/2.0*PI/180.0) * (cast(double)(starty) - (dwin_h/2.0))/(dwin_h/2.0)) ;
        startx=x ;
        starty=y ;
        return(1) ;
      }
    }
    startx=x ;
    starty=y ;
    return(0) ;
  }

  int changeOBJ(int no){
    solidobj=no ;
    RedrawWindow( getHWND(), null, null, RDW_INVALIDATE) ;
    return(0) ;
  }
  
  int mouseRedrawble( int acs_stat, int b, int x, int y ){
    int ret=mouse(acs_stat,b,x,y) ;
    RedrawWindow( getHWND(), null, null, RDW_INVALIDATE) ;
    return(ret) ;
  }
}
