import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.WindowObject;
import std.stdio;
import std.string;
import std.utf;

import SliderBase;

COLORREF RGBA(uint r, uint g, uint b, uint a) {
  return cast(COLORREF)(r | (g << 8) | (b << 16) | (a<<24));
}

class SliderGrad : SliderBase {
  HDC        WinDC;
  HDC        WinDC_back;
  HBITMAP    bmp_back  ;
  BITMAPINFO bi ;
  uint*      bgmap ;

  this(string name, string ClassName) {
    super(name,ClassName) ;
  }

  int wm_destroy(HWND hwnd, WPARAM wp , LPARAM lp){
    DeleteDC(WinDC);
    DeleteDC(WinDC_back);
    DeleteObject(bmp_back);
    return(0) ;
  }
  
  int wm_create(HWND hwnd, WPARAM wp , LPARAM lp){
    bi.bmiHeader.biSize             = BITMAPINFOHEADER.sizeof;
    bi.bmiHeader.biWidth            =  (winW>0) ?  winW :  2 ;
    bi.bmiHeader.biHeight           =  (winH>0) ? -winH : -2 ;
    bi.bmiHeader.biPlanes           = 1;
    bi.bmiHeader.biBitCount         = 32;
    bi.bmiHeader.biCompression      = BI_RGB;
    bi.bmiHeader.biSizeImage        = 0;
    bi.bmiHeader.biXPelsPerMeter    = 0;
    bi.bmiHeader.biYPelsPerMeter    = 0;
    bi.bmiHeader.biClrUsed          = 1;
    bi.bmiHeader.biClrImportant     = 0;

    WinDC=GetDC(hwnd);
    WinDC_back=CreateCompatibleDC(WinDC);
    bmp_back  =CreateDIBSection( WinDC, &bi, DIB_RGB_COLORS, cast(void**)(&bgmap), null, 0 );
    SelectObject(WinDC_back, bmp_back);
    fillGrad() ;
    
    return(0) ;
  }
  
  int wm_paint(HWND hwnd, WPARAM wp , LPARAM lp){
    HDC hdc;
    PAINTSTRUCT ps ;
    HBRUSH COLPTR ;

    COLPTR = CreatePen( PS_SOLID, 1, RGBA(255,255,255,0) ) ;
    hdc = BeginPaint(hwnd, &ps);

    if( slidertype==0 ){
      StretchBlt(hdc, 0, 0, winW, winH, WinDC_back, 0, 0, winW, winH, SRCCOPY);
      SelectObject(hdc, COLPTR) ;
      SetROP2(hdc , R2_XORPEN);
      MoveToEx(hdc, sliderx,    0, null) ;
      LineTo(  hdc, sliderx, winH ) ;
    }else{
      SelectObject(hdc, COLPTR) ;
      StretchBlt(hdc, 0, 0, winW, winH, WinDC_back, 0, 0, winW, winH, SRCCOPY);
      SelectObject(hdc, COLPTR) ;
      SetROP2(hdc , R2_XORPEN);
      MoveToEx(hdc,    0, slidery, null) ;
      LineTo(  hdc, winW, slidery ) ;
    }
    EndPaint(hwnd, &ps);
    DeleteObject(COLPTR) ;
	
    return(0) ;
  }
  
  int wm_size(int w, int h){
    if( bi.bmiHeader.biWidth!=w || bi.bmiHeader.biHeight!=h || bmp_back != null ){
      DeleteDC(WinDC_back);
      DeleteObject(bmp_back);
      bi.bmiHeader.biWidth            =  (w>0) ?  w :  2 ;
      bi.bmiHeader.biHeight           =  (h>0) ? -h : -2 ;
      WinDC_back=CreateCompatibleDC(WinDC);
      bmp_back  =CreateDIBSection( WinDC, &bi, DIB_RGB_COLORS, cast(void**)(&bgmap), null, 0 );
      SelectObject(WinDC_back, bmp_back);
      fillGrad() ;
    }
    return(0) ;
  }

  int fillGrad(){
    int x,y ;
    int r,b,g,a ;
    int rfg,gfg,bfg,afg ;
    int rbg,gbg,bbg,abg ;
    int bgv   ;
    if( bgmap != null ){
      if( slidertype == 0 ){
	for( y=0 ; y<winH ; y++ ){
	  for( x=0 ; x<winW ; x++ ){
	    afg = (colfg>>24) & 0x000000ff ;
	    rfg = (colfg    ) & 0x000000ff ;
	    gfg = (colfg>>8 ) & 0x000000ff ;
	    bfg = (colfg>>16) & 0x000000ff ;
	    
	    abg = (colbg>>24) & 0x000000ff ;
	    rbg = (colbg    ) & 0x000000ff ;
	    gbg = (colbg>>8 ) & 0x000000ff ;
	    bbg = (colbg>>16) & 0x000000ff ;

	    r = ( (rfg * x) + (rbg * (winW-x)) ) /(winW) ;
	    b = ( (bfg * x) + (bbg * (winW-x)) ) /(winW) ;
	    g = ( (gfg * x) + (gbg * (winW-x)) ) /(winW) ;
	    a = ( (afg * x) + (abg * (winW-x)) ) /(winW) ;

	    bgv = ((x/8+y/8)%2)==0 ? 128 : 70 ;
	    r = (r*a + bgv*(255-a)) /255 ;
	    g = (g*a + bgv*(255-a)) /255 ;
	    b = (b*a + bgv*(255-a)) /255 ;
	    bgmap[y*winW+x] = RGBA(b,g,r,255) ;
	  }
	}
      }else{
	for( y=0 ; y<winH ; y++ ){
	  for( x=0 ; x<winW ; x++ ){
	    afg = (colfg>>24) & 0x000000ff ;
	    rfg = (colfg    ) & 0x000000ff ;
	    gfg = (colfg>>8 ) & 0x000000ff ;
	    bfg = (colfg>>16) & 0x000000ff ;
	    
	    abg = (colbg>>24) & 0x000000ff ;
	    rbg = (colbg    ) & 0x000000ff ;
	    gbg = (colbg>>8 ) & 0x000000ff ;
	    bbg = (colbg>>16) & 0x000000ff ;

	    r = ( (rfg * y) + (rbg * (winH-y)) ) /(winH) ;
	    b = ( (bfg * y) + (bbg * (winH-y)) ) /(winH) ;
	    g = ( (gfg * y) + (gbg * (winH-y)) ) /(winH) ;
	    a = ( (afg * y) + (abg * (winH-y)) ) /(winH) ;

	    bgv = ((x/8+y/8)%2)==0 ? 128 : 70 ;
	    r = (r*a + bgv*(255-a)) /255 ;
	    g = (g*a + bgv*(255-a)) /255 ;
	    b = (b*a + bgv*(255-a)) /255 ;
	    bgmap[y*winW+x] = RGBA(b,g,r,255) ;
	  }
	}
      }
    }
    return(0) ;
  }
  
}
