import win32.windows;
import win32.windef;
import win32.commctrl;
import win32frame.WindowObject;
import win32frame.Scrollbar;
import win32frame.TextBox;
import win32frame.UpDownCtrl;
import std.stdio;
import std.string;
import std.utf;
import std.conv;

enum CMD{
  UPDN  =0x301,
  TXTBOX=0x302
} ;
    
class Spin : WindowObject {
  HGLRC     hRC;

  UpDownCtrl udcntl ;
  bool       update_text ;
  
  TextBox   textbox ;
  string    textfmt ;
  double    cur, max, min, step ;
  int       scrW = 12 ;
  
  this(string name, string ClassName) {
    winc.style          = 0;
    winc.cbClsExtra     = 0;
    winc.cbWndExtra     = 0;
    winc.lpszMenuName   = null;
    winc.hCursor        = LoadCursor(null , cast(wchar*)(IDC_ARROW));
    winc.lpszClassName  = toUTFz!(wchar*)(ClassName) ;
    super(name) ;
    this.registwinc() ;

    cur= 0.0 ;
    min= 1.0 ;
    max=-1.0 ;
    step=1.0 ;
    textfmt="%.0f" ;
    update_text=false ;
  }

  int setPosition(int x, int y, int w, int h){
    winX = x ;
    winY = y ;
    winW = w ;
    winH = h ;
    return(0) ;
  }

  int setInfo(double cur, double min, double max, double step){
    this.cur = cur ;
    this.min = min ;
    this.max = max ;
    this.step= step ;
    return(0) ;
  }

  int setValue(double cur){
    if( min<=max ){
      cur = (cur<min) ? min : (cur>max) ? max : cur ;
    }
    if( textbox.getHWND() != null ){
      textbox.setWindowText(format(textfmt,cur)) ;
    }
    this.cur=cur ;
    return(0) ;
  }
  
  double getValue(){
    return(cur) ;
  }

  int setTextFormat(string fmt){
    textfmt=fmt ;
    return(0) ;
  }

  int setTextFont(HFONT hfont){
    if( getHWND() != null ){
      return(textbox.setFont(hfont)) ;
    }
    return(-1) ;
  }
  
  LRESULT proc(HWND hwnd , UINT msg , WPARAM wp , LPARAM lp){
    switch (msg) {
    case WM_DESTROY:
      break ;
      
    case WM_CREATE:
      udcntl = new UpDownCtrl("updowncontrol") ;
      udcntl.setPosition(winW-scrW,winY,scrW,winH) ;
      addObject(udcntl, cast(HMENU)(CMD.UPDN), false, false) ;
      udcntl.setRange(0,100) ;
      udcntl.setPos(50) ;

      int tabstop = getStyle() & WS_TABSTOP ;
      string defaultText = format(textfmt,cur) ;
      textbox= new TextBox(defaultText) ;
      textbox.addExStyle(WS_EX_CLIENTEDGE) ;
      textbox.addStyle(WS_BORDER | ES_RIGHT | ES_AUTOHSCROLL | tabstop) ;
      textbox.setPosition(winX,winY,winW-scrW,winH);
      addObject(textbox, cast(HMENU)(CMD.TXTBOX), false, false) ;

      break ;

    case WM_SIZE :
      if( (udcntl !is null) && (textbox !is null) ){
	MoveWindow(  udcntl.getHWND(), winW-scrW, 0,      scrW,winH,true) ;
       	MoveWindow( textbox.getHWND(),         0, 0, winW-scrW,winH,true) ;
      }
      break ;

    case WM_VSCROLL:
      bool   changed=false ;
      double d=0.0 ;
      if( HIWORD(wp)<50 ){
	d =  step * (50 - HIWORD(wp));
	changed=true ;
      }else if( HIWORD(wp)>50 ){
	d = -step * (HIWORD(wp) -50 );
	changed=true ;
      }
      udcntl.setPos(50) ;
      if( changed==true ){
	SetFocus(null) ;
	SetFocus(textbox.getHWND()) ;
	cur += d ;
	if( min<=max ){
	  cur = (cur<min) ? min : (cur>max) ? max : cur ;
	}
	textbox.setWindowText(format(textfmt,cur)) ;
	sendmsgtoparent( WM_COMMAND, cast(int)(getHMENU()), 0) ;
      }
      break ;

    case WM_SETFOCUS :
      SetFocus(textbox.getHWND()) ;
      sendmsg(textbox.getHWND(), EM_SETSEL, 0,65535) ;
      break ;
      
    case WM_COMMAND:
      switch( LOWORD(wp) ){
      case CMD.TXTBOX:
	if( HIWORD(wp)==EN_KILLFOCUS ){
	  string val=textbox.getWindowText() ;
	  if( isNumeric(val)==true ){
	    cur=to!(double)(val) ;
	    if( min<=max ){
	      cur = (cur<min) ? min : (cur>max) ? max : cur ;
	    }
	  }
	  textbox.setWindowText(format(textfmt,cur)) ;
	  sendmsgtoparent( WM_COMMAND, cast(int)(getHMENU()), 0) ;
	}
	break ;
      default :	break ;
      }
      break ;
      
    default :
      return( DefWindowProc(hwnd , msg , wp , lp) );
    }
    return(0);
  }

}
